#! /usr/bin/env python
###############################################################################
#
#  Script name                  : Logger.py
#
#  Dependencies                 : ComDefs.py
#                    
#  Objective                    : Class to provide logging for BladeCenterUpdates
#                    
#  Written by                   : John R. Hicks
#                     
#  Change History               : who -   when   - what
#                                 JRH - 07/08/04 - Initial revision.
#                                 RKC - 08/01/04 - use getopt instead of optparse
#  
#  Copyright IBM Corporation, 2004
#
###############################################################################

try:
    import sys, getopt, os
    from time import localtime, strftime
    sys.path.append("..")
    import ComDefs, ComFuncs
except KeyboardInterrupt:
    sys.exit(1)


class Logger:
    # The name of the BCU log file
    if os.name == "posix":
        LOGFILENAME = "/tmp/uxbc.log"
    else:
        LOGFILENAME = os.environ['TEMP'] + "\uxbc.log"
    logfile = ""
    
    def __init__(self, params):
        
        """ Constructor that sets the optional global debug flag
        Specify True (default is False) for isDebug to display addtional
        messages when using the functions of this class
        """
        self.params = params
        self.logfile = open(self.LOGFILENAME, 'a')
        self.logfile.write(self.timestamp() + " - " + "Logging started!" + "\n")
        self.logfile.close()
    
    def __del__(self):
        try:
            self.logfile = open(self.LOGFILENAME, 'a')
            self.logfile.write(self.timestamp() + " - " + "Logging stopped!" + "\n")
            self.logfile.close()
        except TypeError:
            print
         
    def showLog(self):
        """ Displays log to user """
        self.logfile = open(self.LOGFILENAME, 'r')
        print self.logfile.read()
        self.logfile.close()
        
    def log(self, message, forceOutput=ComDefs.FALSE):
        """ Writes a message to the log file.
            
            All messages get printed to the log file.
            If in verbose mode or forceOutput is specified the message
            also gets printed to the screen.

        """
        self.logfile = open(self.LOGFILENAME, 'a')
        self.logfile.write(self.timestamp() + " - " + message + "\n")
        self.logfile.close()
        if self.params.debugLevel > 0 or forceOutput == ComDefs.TRUE:
            print message
        
    def timestamp(self):
        """ Returns a timestamp for use in log messages """
        return strftime("%a %m/%d/%Y %I:%M:%S %p", localtime())
    
    def close(self):
        self.__del__()
        
# The last error that occurred
lastError   = ComDefs.SUCCESS
isRead = ComDefs.FALSE
writeval = ""

def usage():
    '''
        Display the Usage Information to the screen
    '''

    print "Usage: Logger.py [options]"
    print "\nOptions:"
    print "  -r, --read           Reads and displays the contents of the log file"
    print "  -w, --write=MESSAGE  Writes the specified MESSAGE to the log file"
    print "  -h, --help, -?       Display this (help) message"
    print "  -v, --verbose        Increase the debug output level by one"
    print "  -V, --version        Display the Logger version number and exit"


def parse_args(opts, params):
    """
       Parse the arguments and options and handle them accordingly
    """

    global isRead, writeval

    if opts == []:
        usage()
        return 1

    for o, a in opts:
        if o in ("-?", "-h", "--help"):
            usage()
            return 1
        elif o in ("-v", "--verbose"):
            params.debugLevel += 1
        elif o in ("-V", "--version"):
            print ComDefs.VERSION
            return 1
        elif o in ("-r", "--read"):
            isRead = ComDefs.TRUE
        elif o in ("-w", "--write"):
            # NOTE: In order to enter a string on the command line,
            #       it must be surrounded by quotes
            writeval = a
        else:
            usage()
            return 1

    return 0


def main():
    """ Main function used to run this script in standalone mode """
    # Define using a global variable
    global lasterror
    
    # Instantiate the class
    params = ComDefs.parameters()

    # Parse command line options with getopt()
    try:
        opts, args = getopt.getopt(sys.argv[1:], "?hvVrw:",
             ["?", "help", "verbose", "version", "read", "write="])

    except getopt.GetoptError:
        # Print usage information and exit
        usage()
        sys.exit(2)
    
    if parse_args(opts, params):
        return 0
    
    logger = Logger(params)

    if isRead:
        logger.showLog()

    if writeval:
        logger.log(writeval)
    
# This get's invoked if they run this script directly
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        ComFuncs.keyboardISR(logger)
