#! /usr/bin/env python
###############################################################################
#
#  Script name                  : MMComunications.py
#
#  Dependencies                 : ComDefs.py
#                    
#  Objective                    : Communication class for connecting to the management
#                                 port (6090) of the Management Module and running dot
#                                 commands
#                    
#  Written by                   : Mike Nolterieke
#                     
#  Change History               : who -   when   - what
#                                 MHN - 07/01/04 - Initial revision.
#                                 RKC - 08/01/04 - added logging
#  
#  Copyright IBM Corporation, 2004
#
###############################################################################
""" Communication class for connecting to the management port (6090) of the
    Management Module and running dot commands
"""
try:
    import socket, sys, time, string
    sys.path.append("..")
    import ComDefs
except KeyboardInterrupt:
    sys.exit(1)

logger = ""

class MMCom:
    
    def __init__(self, isDebug=ComDefs.FALSE):
        """ Constructor that sets the optional global debug flag
        Specify True (default is False) for isDebug to display addtional
        messages when using the functions of this class
        """
        # Check Python version, and exit if we are too low
        version = string.split(string.split(sys.version)[0], ".")
        if map(int, version) < [2, 3]:
            print "This program requires Python version 2.3 or newer."
            print
            sys.exit()
        self.logger = logger
        self.isDebug = isDebug

    def open(self, doConnect, host, port=6090, timeOut=5, userName="USERID", password="PASSW0RD"):
        """ Configure and open the connection to the MM specified by the host parameter
        using the specified port (6090 by default) and the specified timeout value
        (5 seconds by default)
        
        If doConnect is True, this function will call the sendUserName function using the
        specified userName (USERID by default) and the sendPassword function using the
        specified password (PASSW0RD by default) to login into the specified MM
         
        Return Codes:
            ComDefs.SUCCESS
            ComDefs.MM_SOCKET_ERR
            ComDefs.MM_BAD_IP
            ComDefs.MM_UNKNOWN_ERR
            ComDefs.MM_SOCKET_TIMEOUT
            Any return codes possible from the sendUserName and sendPassword functions
        """

        # Check to verify that the username is between 1 and 15 characters
        if len(userName) < 1 or len(userName) > 15:
            self.logger.log("Username is invalid.  Usernames must contain between 1 and 15 characters.")
            return ComDefs.MM_INVALID_LOGIN
            
        # Check to verify that the password is between 5 and 15 characters if a password
        # is specified at all.  The MM spec allows for zero-length passwords.
        if len(password) > 0:
            if len(password) < 5 or len(password) > 15:
                self.logger.log("Password is invalid.  Empty passwords are permitted, but if a password is specified,")
                self.logger.log("it must contain between 5 and 15 characters.")
                return ComDefs.MM_INVALID_LOGIN

        self.host = host
        self.port = port

        self.logger.log("Opening connection to Management Module %s on port %d." % (self.host, self.port))

        # Create the socket and handle any errors
        try:
            # Configure the socket
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            # Set the default timeout to 5 seconds
            socket.setdefaulttimeout(5)
        except:
            self.logger.log("Unknown socket error %s" % sys.exc_info()[0])
            self.logger.log("Data was: %s" % sys.exc_info()[1])
            return ComDefs.MM_SOCKET_ERR

        # Connect to the MM and handle any errors
        try:
            self.sock.connect((self.host, self.port))
        except socket.gaierror:
            self.logger.log("Bad IP error")
            return ComDefs.MM_BAD_IP
        except socket.error:
            #self.logger.log("Socket error")
            return ComDefs.MM_SOCKET_ERR
        except socket.timeout:
            #self.logger.log("Socket connect timeout.")
            return ComDefs.MM_SOCKET_TIMEOUT
        except:
            self.logger.log("Unknown socket error %s" % sys.exc_info()[0])
            self.logger.log("Data was: %s" % sys.exc_info()[1])
            return ComDefs.MM_UNKNOWN_ERR

        # If they do not want to login yet, return success
        if doConnect == ComDefs.FALSE:
            return ComDefs.SUCCESS

        # If they want to login...
        # Send the specified user name and get the encryption keys back
        keys = self.sendUserName(userName)
        # Make sure we got the encryption keys
        if keys == ComDefs.MM_INVALID_KEYS:
            return keys
        # Send the specified password using the encryption keys
        rc = self.sendPassword(password, keys);
        # Return the return code from sendPassword
        return rc

    def close(self):
        """ Closes the communication socket"""
        self.logger.log("Closing communications socket.")
        self.sock.close()
        self.sock = 0

    def sendUserName(self, userName):
        """ Send the specified user name to log into the MM and receive and return the encryption keys
        Return Codes:
            ComDefs.MM_INVALID_KEYS"""
        self.logger.log("Sending username to Management Module.")
        self.sock.send('SPREMOTE'+ userName + chr(0))

        self.logger.log("Receiving encryption keys.")
        keys = self.sock.recv( 40 )
        if len(keys) != 2:
            self.logger.log("Encryption keys not received.")
            return ComDefs.MM_INVALID_KEYS
        return keys

    def sendPassword(self, password, keys):
        """ Send the encrypted password to the MM using the specified password and encryption keys 
        Return Codes: 
            ComDefs.SUCCESS
            ComDefs.MM_INVALID_LOGIN
            ComDefs.MM_RETRY_LIMIT_EXCEEDED
        """
        keybyte0 = (ord(keys[0]))
        keybyte1 = (ord(keys[1]))

        # pw = password exclusive OR'd with 1st keybyte then added with 2nd keybyte
        pw = ''
        i = 0
        while i < len(password):
            pw += chr(((ord(password[i]) ^ keybyte0) + keybyte1) & 0xFF)
            i += 1

        # Send the encrypted password
        self.logger.log("Sending encrypted password to Management Module.")
        self.sock.send(pw + chr(0))

        # Receive the acknowledgement back from the login attempt and check for errors
        ack = self.sock.recv(80)
        if ack == "BADLOGIN\0":
            self.logger.log("An invalid user ID or password was specified!")
            return ComDefs.MM_INVALID_LOGIN
        elif ack == "RETRYLIMIT\0":
            self.logger.log("The retry limit has been exceeded!")
            return ComDefs.MM_RETRY_LIMIT_EXCEEDED
        else:
            self.logger.log("Login was successful!")
            return ComDefs.SUCCESS

        return ComDefs.SUCCESS

    def sendCmd(self, cmdType, cmd, data):
        """ Sends the specified dot cmd and data to the opened connection, waits
        for response, and validates that the received response is for the issued cmd.
        The response data can then be obtained via the getResponse() method.

        The command type (cmdType) must be one of the following:
            R   = Read
            RN  = Read Next
            W   = Write
            WN  = Write Next

        The command (cmd) must be in hex format (\\x07\\x04 to reboot the MM)

        Return Codes:
            ComDefs.SUCCESS
            ComDefs.MM_UNKNOWN_CMD_TYPE
            ComDefs.MM_UNKNOWN_ERR
            ComDefs.MM_INVALID_RESPONSE
            ComDefs.MM_LENGTH_MISMATCH
            ComDefs.MM_COMMAND_MISMATCH
            ComDefs.MM_SEND_FAILURE
        """

        # Print out command in hex if we are in debug mode
        if self.isDebug > 2:
            # Convert the command
            self.cmdHex = ''
            indx = 0
            while indx < len(cmd):
                self.cmdHex += "%X" % ord(cmd[indx])
                indx += 1
                # Only add the period if it is not the last one
                if indx < len(cmd):
                    self.cmdHex += "."

            self.logger.log('Sending command (%s) %s' % (cmdType, self.cmdHex))

        # Get the length of the command
        self.cmdLen = len(cmd)
        # Determine the type of command and convert it to an int
        # Write
        if cmdType == 'W':
            cmdType = 0
        # Write Next
        elif cmdType == 'WN':
            cmdType = 1
        # Read
        elif cmdType == 'R':
            cmdType = 2
        # Read Next
        elif cmdType == 'RN':
            cmdType = 3
        # Unknown
        else:
            self.logger.log("Invalid command type.")
            return ComDefs.MM_UNKNOWN_CMD_TYPE
        # Convert the command type to hex
        cmdType = chr(cmdType)

        # Set the default return status to success
        self.rtnstatus = ComDefs.SUCCESS

        count = 0
        # We will retry the command until we succeed or reach MAX_MM_SEND_RETRIES
        while count < ComDefs.MAX_MM_SEND_RETRIES:
            senderror = ComDefs.FALSE
            # Send the command
            try:
                self.sock.send(cmdType + chr(self.cmdLen) + chr(len(data)) + (chr(0)*3) + cmd + data + chr(0))
            except socket.timeout:
                #self.logger.log("Socket send() timeout.")
                return ComDefs.MM_SOCKET_TIMEOUT
            except socket.error:
                #self.logger.log("Socket error")
                return ComDefs.MM_SOCKET_ERR
            except:
                self.logger.log("Unknown error %s sending the command" % sys.exc_info()[0])
                self.logger.log("Data was: %s" % sys.exc_info()[1])
                return ComDefs.MM_UNKNOWN_ERR

            if senderror == ComDefs.FALSE:
            # Skip the response data when restarting
                if (cmd == '\x07\x04'):
                    self.logger.log("Reboot and switchover command was issued, skipping response.")
                    return ComDefs.SUCCESS
                if (cmd == '\x07\x04\x01'):
                    self.logger.log("Reboot command was issued, skipping response.")
                    return ComDefs.SUCCESS

                # Get the response
                try:
                    self.response = self.sock.recv(1500)
                # Handle errors
                except socket.timeout:
                    #self.logger.log("Socket recv() timeout.")
                    return ComDefs.MM_SOCKET_TIMEOUT
                except socket.error:
                    #self.logger.log("Socket error")
                    return ComDefs.MM_SOCKET_ERR
                except:
                    self.logger.log("Unknown error %s getting response" % sys.exc_info()[0])
                    self.logger.log("Data was: %s" % sys.exc_info()[1])
                    return ComDefs.MM_UNKNOWN_ERR

                # Validate the response if we were able to get it
                if senderror == ComDefs.FALSE:
                    # Invalid responses
                    if (len(self.response) == 0):
                        self.logger.log("Response is empty.")
                        self.rtnstatus = ComDefs.MM_INVALID_RESPONSE
                        senderror = ComDefs.TRUE
                        #return ComDefs.MM_INVALID_RESPONSE
                    elif self.response[0] != chr(4):
                        self.logger.log("Response is not valid")
                        self.rtnstatus = ComDefs.MM_INVALID_RESPONSE
                        senderror = ComDefs.TRUE
                        #return ComDefs.MM_INVALID_RESPONSE
                    # Length mismatch
                    elif self.response[1] != chr(self.cmdLen):
                        self.logger.log("Response length does not match command length")
                        self.rtnstatus = ComDefs.MM_LENGTH_MISMATCH
                        senderror = ComDefs.TRUE
                        #return ComDefs.MM_LENGTH_MISMATCH
                    # Command mismatch
                    elif (self.response[6:6+self.cmdLen] != cmd[0:self.cmdLen]):
                        self.logger.log("Response command does not match sent command")
                        self.rtnstatus = ComDefs.MM_COMMAND_MISMATCH
                        senderror = ComDefs.TRUE
                        #return ComDefs.MM_COMMAND_MISMATCH

                    # Check for a successful command
                    if senderror == ComDefs.FALSE:
                        self.rtnstatus = ord(self.response[4])
                        if self.rtnstatus == ComDefs.SUCCESS:
                            break
                        # If the command is not supported, there is no reason to retry
                        if self.rtnstatus == 2:
                            self.logger.log("Command is not supported.")
                            return ComDefs.MM_SEND_FAILURE
                        else:
                            senderror = ComDefs.TRUE
                    
            count += 1

            if count < ComDefs.MAX_MM_SEND_RETRIES:
                if self.isDebug > 2:
                    self.logger.log("Response status was %d" % self.rtnstatus)
                    self.logger.log("Retrying SP command #%d of %d" % (count, ComDefs.MAX_MM_SEND_RETRIES))
                # Sleep and then retry
                time.sleep(0.2)
            else:
                return ComDefs.MM_SEND_FAILURE
        
        # If we tried the command too many times and did not succeed
        if senderror == ComDefs.TRUE:
            self.logger.log("Failure sending command!")
            return ComDefs.MM_SEND_FAILURE

        # Print out response in hex if we are in debug mode
        if self.isDebug > 3:
            # Convert the response
            self.datalen  = ord(self.response[3]) * 256
            self.datalen += ord(self.response[2])
            self.datahex = ''
            indx = 0
            indx2 = 6 + self.cmdLen
            while indx < self.datalen:
                if indx2 >= len(self.response):
                    break
                self.datahex += "%02X" % ord(self.response[indx2])
                indx += 1
                indx2 += 1
                # Only add the comma if it is not the last one
                if indx < self.datalen:
                    self.datahex += ","

            self.logger.log('Hex response data for dot command %s is %s' % (self.cmdHex, self.datahex))

        # Get the length of the response
        indx2 = 6 + self.cmdLen
        # Put the response in a global variable
        self.responseData = self.response[indx2:]
        return ComDefs.SUCCESS

    def getResponse(self):
        """  Returns the last dot command response data.
        """
        return self.responseData

    def sendPacket(self, packet):
        """ Sends the specified packet file to the opened connection, waits
        for response, and validates that the received response is for the issued cmd.

        Return Codes:
            ComDefs.SUCCESS
            ComDefs.MM_INVALID_RESPONSE
            ComDefs.MM_LENGTH_MISMATCH
            ComDefs.MM_COMMAND_MISMATCH
            ComDefs.MM_SEND_FAILURE
        """

        self.cmdType = packet[0]
        self.cmdLen  = ord(packet[1])

        # Set the default return status to success
        self.rtnstatus = ComDefs.SUCCESS

        count = 0
        # We will retry the command until we succeed or reach MAX_MM_SENDPKT_RETRIES
        while count < ComDefs.MAX_MM_SENDPKT_RETRIES:
            senderror = ComDefs.FALSE
            # Send the command
            try:
                self.sock.send(packet)
            except socket.timeout:
                #self.logger.log("Socket send() timeout.")
                return ComDefs.MM_SOCKET_TIMEOUT
            except socket.error:
                #self.logger.log("Socket error")
                return ComDefs.MM_SOCKET_ERR
            except:
                self.logger.log("Unknown error %s sending the packet" % sys.exc_info()[0])
                self.logger.log("Data was: %s" % sys.exc_info()[1])
                return ComDefs.MM_UNKNOWN_ERR
        
            if senderror == ComDefs.FALSE:
                # Get the response
                try:
                    self.response = self.sock.recv(200)
                # Handle errors
                except socket.timeout:
                    #self.logger.log("Socket recv() timeout.")
                    return ComDefs.MM_SOCKET_TIMEOUT
                except socket.error:
                    #self.logger.log("Socket error")
                    return ComDefs.MM_SOCKET_ERR
                except:
                    self.logger.log("Unknown error %s getting response" % sys.exc_info()[0])
                    self.logger.log("Data was: %s" % sys.exc_info()[1])
                    return ComDefs.MM_UNKNOWN_ERR

                # Validate the response if we were able to get it
                if senderror == ComDefs.FALSE:
                    # Invalid responses
                    if (len(self.response) == 0):
                        self.logger.log("Response is empty.")
                        self.rtnstatus = ComDefs.MM_INVALID_RESPONSE
                        senderror = ComDefs.TRUE
                        #return ComDefs.MM_INVALID_RESPONSE
                    elif self.response[0] != chr(4):
                        self.logger.log("Response is not valid")
                        self.rtnstatus = ComDefs.MM_INVALID_RESPONSE
                        senderror = ComDefs.TRUE
                        #return ComDefs.MM_INVALID_RESPONSE
                    # Length mismatch
                    elif self.response[1] != chr(self.cmdLen):
                        self.logger.log("Response length does not match command length")
                        self.rtnstatus = ComDefs.MM_LENGTH_MISMATCH
                        senderror = ComDefs.TRUE
                        #return ComDefs.MM_LENGTH_MISMATCH
                    # Command mismatch
                    elif (self.response[6:6+self.cmdLen] != packet[6:6+self.cmdLen]):
                        self.logger.log("Response command does not match sent command")
                        self.rtnstatus = ComDefs.MM_COMMAND_MISMATCH
                        senderror = ComDefs.TRUE
                        #return ComDefs.MM_COMMAND_MISMATCH

                    # Check for a successful command
                    if senderror == ComDefs.FALSE:
                        self.rtnstatus = ord(self.response[4])
                        if self.rtnstatus == ComDefs.SUCCESS:
                            break
                        else:
                            senderror = ComDefs.TRUE
                    
            count += 1

            if count < ComDefs.MAX_MM_SENDPKT_RETRIES:
                self.logger.log("Response status was %d" % self.rtnstatus)
                self.logger.log("Retrying SP command #%d of %d" % (count, ComDefs.MAX_MM_SENDPKT_RETRIES))
                # Sleep and then retry
                time.sleep(0.2)
            else:
                return ComDefs.MM_SEND_FAILURE
        
        # If we tried the command too many times and did not succeed
        if senderror == ComDefs.TRUE:
            self.logger.log("Failure sending packet!")
            return ComDefs.MM_SEND_FAILURE

        return ComDefs.SUCCESS
