#! /usr/bin/env python
#DLNK
###############################################################################
#
#  Script name                  : IBMSMFlash.py
#
#  Dependencies                 : ComDefs.py
#                                 ComFuncs.py
#                                 Logger.py
#                    
#  Objective                    : Script used to login to a IBM ESM 
#                                 using telnet and update the firmware via a 
#                                 TFTP server.
#                    
#  Written by                   : Robert K. Christner
#                     
#  Change History               : who -   when   - what
#                                 RKC - 07/08/04 - Initial revision.
#  
#  Copyright IBM Corporation, 2004
#
###############################################################################
"""
   Script used to login to a IBM switch over telnet and update the
   firmware via a TFTP server.
"""
try:
    import time, sys, os, re, getopt, telnetlib, socket
    sys.path.append("..")
    from Common import ComDefs, ComFuncs
    from Common.Logger import Logger
except KeyboardInterrupt:
    sys.exit(1)

# Instantiate Globals
params = ""
logger = ""
lastError = ComDefs.SUCCESS
ipAddr = ComDefs.FALSE
isNOTstandalone = ComDefs.TRUE
cleanExit = ComDefs.TRUE

def usage():
    '''
        Display the Usage Information to the screen
    '''

    print "Usage: IBMSMFlash.py [options]"
    print "\nOptions:"
    print "  -i, --ipaddress=IPADDR   The hostname or dotted IP address of the I/O Module"
    print "  -s, --tftpserver=IPADDR  The dotted IP address of the TFTP server hosting the"
    print "                           new firmware image"
    print "  -n, --imagename=FILE     The path and name of the I/O Module firmware image"
    print "                           file on the TFTP server"
    print "  -u, --username=USERNAME  An administrator user name used to log into the I/O"
    print "                           Module"
    print "  -p, --password=PASSWORD  The password of the administrator user name used to"
    print "                           log into the I/O Module"
    print "  -h, --help, -?           Display this (help) message"
    print "  -v, --verbose            Increase the debug output level by one"
    print "  -V, --version            Display the IBMSMFlash version number and exit"


def parse_args(opts):
    """
       Parse the arguments and options and handle them accordingly
    """

    global params, ipAddr, cleanExit

    for o, a in opts:
        if o in ("-?", "-h", "--help"):
            usage()
            return cleanExit
        elif o in ("-v", "--verbose"):
            params.debugLevel += 1
        elif o in ("-V", "--version"):
            print ComDefs.VERSION
            return cleanExit
        elif o in ("-i", "--ipaddress"):
            ipAddr = a
        elif o in ("-s", "--tftpserver"):
            params.tftpIPAddr = a
        elif o in ("-n", "--imagename"):
            params.io1Filename1 = a
        elif o in ("-u", "--username"):
            params.io1UserName = a
        elif o in ("-p", "--password"):
            params.io1Password = a
        else:
            usage()
            return cleanExit

    return ComDefs.SUCCESS

def verify_args():
    """
       Verify that the required arguments have been supplied.
    """

    global params, ipAddr, cleanExit

    if ipAddr == "":
        print "The hostname or dotted IP address of the I/O Module was not specified."
        usage()
        return cleanExit
    if params.tftpIPAddr == "":
        print "The dotted IP address of the TFTP server was not specified."
        usage()
        return cleanExit
    if params.io1Filename1 == "":
        print "The path and name of the I/O Module firmware image file on the TFTP"
        print "server was not specified."
        usage()
        return cleanExit

    return ComDefs.SUCCESS


def connect(telnet, ioIPAddr, userName, password):
    """
        Connect and authenticate to the switch.
    """

    global params, logger, isNOTstandalone

    try:
        telnet.open(ioIPAddr)
        logger.log("Connecting to %s" % ioIPAddr)
    except socket.error:
        logger.log("Socket error! Connection refused or timed out to (%s)." % ioIPAddr)
        lastError = ComDefs.TN_SOCKET_ERR
        return lastError
    except KeyboardInterrupt:
        sys.exit(0)
    # look for "Username" and enter credentials
    reList = []
    reList.append(re.compile('Username'))
    reList.append(re.compile('Too\ many\ users'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    elif index == 1:
        logger.log("Error! Too many users, cannot connect to I/O Module.")
        telnet.close()
        lastError = ComDefs.TN_INVALID_RESPONSE
        return lastError
    logger.log("Connected to Ethernet Switch Module at (%s)." % ioIPAddr)
    telnet.write('%s\t%s\r' % (userName, password))
    logger.log("Sending Username and Password.")
    # credentials are sent, continue to the next screen
    reList = []
    reList.append(re.compile('Utilities'))
    reList.append(re.compile('Access\ denied'))
    (index,mo,txt) = telnet.expect(reList)
    if index == 1:
        logger.log("Error! Username or Password not accepted.")
        telnet.close()
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    logger.log("Username and Password accepted.")

    lastError = ComDefs.SUCCESS
    return lastError


def update_firmware(telnet, ioIPAddr, userName, password, filename):
    """
        Flash the firmware
    """

    global params, logger, isNOTstandalone

    # open a telnet session with the switch
    lastError = connect(telnet, ioIPAddr, userName, password)
    if lastError != ComDefs.SUCCESS:
        return lastError

    # credentials are accepted, continue to firmware upgrade
    telnet.write('\t\t\t\t\t\t\r')
    telnet.read_until('Upgrade')
    telnet.write('\r')
    telnet.read_until('Server IP')
    # enter TFTP server IP and update image filename
    # the only way to clear the input fields is with a series of \del
    telnet.write('\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f%s\t'\
                  % params.tftpIPAddr)
    telnet.write('\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f')
    telnet.write('\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f')
    telnet.write('\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f')
    telnet.write('\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f')
    telnet.write('\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f\x7f%s\t\r'\
                  % filename)
    logger.log("Entering update filename (%s) and TFTP server IP (%s)."\
                % (filename, params.tftpIPAddr))
    reList = []
    reList.append(re.compile('NVRAM'))
    reList.append(re.compile('Yes'))
    reList.append(re.compile('\/n'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Timeout error while entering TFTP server IP and update filename.")
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    elif index == 0:
        telnet.write('\r')
    elif index == 1:
        telnet.write('\r')
    elif index == 2:
        telnet.write('y')
    # look for the download to start or an error
    logger.log("Downloading Ethernet Switch Module firmware image from TFTP server.", ComDefs.TRUE)
    # wait for download to complete or an error to occur
    reList = []
    #reList.append(re.compile('(1?[0-9]?[0-9]%)'))
    reList.append(re.compile('%'))
    reList.append(re.compile('rasing'))
    reList.append(re.compile('ait'))
    reList.append(re.compile('found'))
    reList.append(re.compile('Found'))
    reList.append(re.compile('issing'))
    reList.append(re.compile('irectory'))
    reList.append(re.compile('ermission'))
    reList.append(re.compile('iolation'))
    while ComDefs.TRUE:
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
        if index == -1:
            print
            logger.log("Error! Telnet session exited abnormally.")
            telnet.close()
            lastError = ComDefs.GENERAL_UPDATE_ERROR
            return lastError
        elif index == 0:
            sys.stdout.write(".")
            sys.stdout.flush()
        elif index == 1:
            print
            logger.log("Erasing FLASH Memory.", ComDefs.TRUE)
            break
        elif index == 2:
            print
            logger.log("Error! Lost connection with TFTP Server.")
            telnet.close()
            lastError = ComDefs.TF_LOST_CONNECTION
            return lastError
        elif index == 3:
            print
            logger.log("Error! Update file not found.")
            telnet.close()
            lastError = ComDefs.IO_FIRMWARE_IMAGE_NOT_FOUND
            return lastError
        elif index == 4:
            print
            logger.log("Error! TFTP server not found.")
            telnet.close()
            lastError = ComDefs.TF_SERVER_NOT_FOUND
            return lastError
        elif index == 5:
            print
            logger.log("Error! The image name that was specified (%s) is not valid." % filename)
            telnet.close()
            lastError = ComDefs.IO_INVALID_IMAGE_FILE
            return lastError
        elif index == 6:
            print
            logger.log("Error! The image name that was specified (%s) is a directory." % filename)
            telnet.close()
            lastError = ComDefs.IO_FIRMWARE_IMAGE_NOT_FOUND
            return lastError
        elif index == 7:
            print
            logger.log("Error! Permission denied to image file on TFTP server.")
            telnet.close()
            lastError = ComDefs.IO_FIRMWARE_IMAGE_NOT_FOUND
            return lastError
        elif index == 8:
            print
            logger.log("Error! Access violation on TFTP server.")
            telnet.close()
            lastError = ComDefs.IO_FIRMWARE_IMAGE_NOT_FOUND
            return lastError

    # Downloading is complete, now writing
    reList = []
    reList.append(re.compile('riting'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Warning! Update completion could not be verified.", ComDefs.TRUE)
        telnet.close()
        lastError = ComDefs.WARNING
        return lastError
    logger.log("Writing FLASH Memory.", ComDefs.TRUE)

    # Writing is complete, now wait for update to finish
    reList = []
    reList.append(re.compile('plete'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Warning! Update completion could not be verified.", ComDefs.TRUE)
        telnet.close()
        lastError = ComDefs.WARNING
        return lastError
    if isNOTstandalone == ComDefs.FALSE:
        logger.log("Ethernet Switch Module firmware succesfully updated.", ComDefs.TRUE)
    telnet.close()
    lastError = ComDefs.SUCCESS
    return lastError


def flash_switch(ioIPAddr, ioBay=1):
    """
       Flash the IBM ESM via interaction with the telnet menu interface.
    """

    global params, logger, isNOTstandalone, lastError

    if isNOTstandalone == ComDefs.FALSE:
        logger.log("Updating Ethernet Switch Module firmware.", ComDefs.TRUE)

    # Get Variables from params class
    if ioBay == 1:
        userName = params.io1UserName
        password = params.io1Password
        filename = params.io1Filename1
    elif ioBay == 2:
        userName = params.io2UserName
        password = params.io2Password
        filename = params.io2Filename1
    elif ioBay == 3:
        userName = params.io3UserName
        password = params.io3Password
        filename = params.io3Filename1
    elif ioBay == 4:
        userName = params.io4UserName
        password = params.io4Password
        filename = params.io4Filename1

    if userName == "":
        logger.log("Error! Username was not supplied.")
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    if password == "":
        logger.log("Error! Password was not supplied.")
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    if filename == "":
        logger.log("Error! Firmware image filename was not supplied.")
        lastError = ComDefs.IO_INVALID_IMAGE_FILE
        return lastError

    # check to make sure tftpIPAddr is valid
    ipValid = re.compile('(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){2}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)')
    ipValidStr = ipValid.match(params.tftpIPAddr)
    if ipValidStr == None:
        logger.log("Error! Invalid TFTP Server IP address (%s)" % params.tftpIPAddr)
        lastError = ComDefs.TF_INVALID_IP
        return lastError

    telnet = telnetlib.Telnet()

    # flash the firmware
    lastError = update_firmware(telnet, ioIPAddr, userName, password, filename)
    if lastError != ComDefs.SUCCESS:
        return lastError

    logger.log("Ethernet Switch Module is restarting.", ComDefs.TRUE)
    if isNOTstandalone == ComDefs.FALSE:
        logger.log("Success! Ethernet Switch Module firmware update complete.", ComDefs.TRUE)

    return lastError


def main():
    """ Main function used to run this script in standalone mode"""

    # Global Variables
    global params, logger, ipAddr, isNOTstandalone, lastError

    params = ComDefs.parameters()
    isNOTstandalone = ComDefs.FALSE

    # Parse command line options with getopt()
    try:
        opts, args = getopt.getopt(sys.argv[1:], "?hvVi:s:n:u:p:",
             ["?", "help", "verbose", "version", 
              "ipaddress=", "tftpserver=", "imagename=", "username=", "password="])

    except getopt.GetoptError:
        # Print usage information and exit
        usage()
        sys.exit(0)
    
    lastError = parse_args(opts)
    if lastError != ComDefs.SUCCESS:
        return lastError

    lastError = verify_args()
    if lastError != ComDefs.SUCCESS:
        return lastError
        
    logger = Logger(params)
    rc = flash_switch(ipAddr)
    if rc != ComDefs.SUCCESS:
        logger.log(ComDefs.GetLastError(rc), ComDefs.TRUE)
    
# This get's invoked if they run this script directly
if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        ComFuncs.keyboardISR(logger)
