#! /usr/bin/env python
#Qlgc
###############################################################################
#
#  Script name                  : QLogicSMFlash.py
#
#  Dependencies                 : ComDefs.py
#                                 ComFuncs.py
#                                 Logger.py
#                    
#  Objective                    : Script used to login to a QLogic switch module 
#                                 and update the firmware via a TFTP server.
#                    
#  Written by                   : John R. Hicks
#                     
#  Change History               : who -   when   - what
#                                 JRH - 08/24/2004 - Initial revision.
#  
#  Copyright IBM Corporation, 2004
#
###############################################################################
"""
    Script used to login to a QLogic switch module 
    and update the firmware via a TFTP server.
"""

try:
    import time, sys, os, os.path, re, getopt, telnetlib, socket
    sys.path.append("..")
    from Common import ComDefs, ComFuncs
    from Common.Logger import Logger
except KeyboardInterrupt:
    sys.exit(1)

# Instantiate Globals
params = ""
logger = ""
lastError = ComDefs.SUCCESS
cleanExit = ComDefs.TRUE
ipAddr = ""
standalone = ComDefs.FALSE


def usage():
    '''
        Display the Usage Information to the screen
    '''

    print "Usage: QLogicSMFlash.py [options]"
    print "\nOptions:"
    print "  -i, --ipaddress=IPADDR      The hostname or dotted IP address of the I/O"
    print "                              Module"
    print "  -u, --username=USERNAME     An administrator user name used to log into the"
    print "                              I/O Module"
    print "  -p, --password=PASSWORD     The password of the administrator user name used"
    print "                              to log into the I/O Module"
    print "  -s, --ftpserver=IPADDR      The dotted IP address of the FTP server hosting"
    print "                              the new firmware image"
    print "  -m, --ftpusername=USERNAME  An user name used to log into the FTP server"
    print "  -w, --ftppassword=PASSWORD  The password of the user name used to"
    print "                              log into the FTP server"
    print "  -n, --imagename=FILE        The path and name of the I/O Module firmware image"
    print "                              file on the FTP server"
    print "  -h, --help, -?              Display this (help) message"
    print "  -v, --verbose               Increase the debug output level by one"
    print "  -V, --version               Display the QLogicSMFlash version number and exit"


def parse_args(opts):
    """
       Parse the arguments and options and handle them accordingly
    """

    global params, ipAddr, cleanExit

    for o, a in opts:
        if o in ("-?", "-h", "--help"):
            usage()
            return cleanExit
        elif o in ("-v", "--verbose"):
            params.debugLevel += 1
        elif o in ("-V", "--version"):
            print ComDefs.VERSION
            return cleanExit
        elif o in ("-i", "--ipaddress"):
            ipAddr = a
        elif o in ("-u", "--username"):
            params.io1UserName = a
        elif o in ("-p", "--password"):
            params.io1Password = a
        elif o in ("-s", "--ftpserver"):
            params.ftpIPAddress = a
        elif o in ("-m", "--ftpusername"):
            params.ftpUserName = a
        elif o in ("-w", "--ftppassword"):
            params.ftpPassword = a
        elif o in ("-n", "--imagename"):
            params.io1Filename1 = a
        else:
            usage()
            return cleanExit

    return ComDefs.SUCCESS

def verify_args():
    """
       Verify that the required arguments have been supplied.
    """

    global params, ipAddr, cleanExit

    if ipAddr == "":
        print "The hostname or dotted IP address of the I/O Module was not specified."
        usage()
        return cleanExit
    if params.ftpIPAddress == "":
        print "The dotted IP address of the FTP server was not specified."
        usage()
        return cleanExit
    if params.io1Filename1 == "":
        print "The path and name of the I/O Module firmware image file on the TFTP"
        print "server was not specified."
        usage()
        return cleanExit

    return ComDefs.SUCCESS

def flash_switch(ioIPAddr, ioBay=1):
    """ Flash the QLogic switch via interaction with the telnet interface. """
    global logger, standalone, params, lastError
    
    if standalone:
        logger.log("Updating Fibre Channel Switch Module firmware.", ComDefs.TRUE)

    # Get Variables from params class
    if ioBay == 1:
        userName = params.io1UserName
        password = params.io1Password
        filename = params.io1Filename1
    elif ioBay == 2:
        userName = params.io2UserName
        password = params.io2Password
        filename = params.io2Filename1
    elif ioBay == 3:
        userName = params.io3UserName
        password = params.io3Password
        filename = params.io3Filename1
    elif ioBay == 4:
        userName = params.io4UserName
        password = params.io4Password
        filename = params.io4Filename1
    
    # check a couple params to make sure
    if params.ftpIPAddress == "":
        logger.log("Error! The IP Address of the ftp server was not specified.")
        lastError = ComDefs.FTP_BAD_IP
        return lastError
    if params.ftpUserName == "":
        logger.log("Error! The Username for the ftp server was not specified.")
        lastError = ComDefs.FTP_INVALID_LOGIN
        return lastError
    if userName == "":
        logger.log("Error! Username was not supplied.")
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    if password == "":
        logger.log("Error! Password was not supplied.")
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    if filename == "":
        logger.log("Error! Firmware image filename was not supplied.")
        lastError = ComDefs.IO_INVALID_IMAGE_FILE
        return lastError

    
    # open a telnet session with the switch
    telnet = telnetlib.Telnet()
    try:
        telnet.open(ioIPAddr)
#        logger.log("Connecting to %s" % ioIPAddr)
    except socket.error:
        logger.log("Socket error! Connection refused or timed out.")
        lastError = ComDefs.TN_SOCKET_ERR
        return lastError
    except KeyboardInterrupt:
        sys.exit(0)
    # look for "login:" and enter credentials
    reList = []
    reList.append(re.compile('ogin'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message received:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    elif index == 1:
        logger.log("Error! Too many users, cannot connect to I/O Module.")
        telnet.close()
        lastError = ComDefs.TN_INVALID_RESPONSE
        return lastError
    logger.log("Connected to Fibre Channel Switch Module at (%s)." % ioIPAddr)
    #telnet.write('%s\t%s\r' % (userName, password))
    #logger.log("Sending Username and Password.")
    telnet.write('%s\r' % userName)
    logger.log("Sending Username.")
    # credentials are sent, continue to the next screen
    reList = []
    reList.append(re.compile('Password'))
    reList.append(re.compile('incorrect'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == 1:
        logger.log("Error! Incorrect username (%s)." % userName)
        telnet.close()
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    elif index == -1:
        logger.log("Error! An unexpected timeout has occurred.")
        logger.log("Last message received:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    telnet.write('%s\r' % password)
    logger.log("Password sent.")
    # credentials are sent, continue to the next screen
    reList = []
    reList.append(re.compile('%s>' % userName))
    reList.append(re.compile('incorrect'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == 1:
        logger.log("Error! Password not accepted.")
        telnet.close()
        lastError = ComDefs.TN_INVALID_LOGIN
        return lastError
    logger.log("Username and Password accepted.")
    
    # credentials are accepted, continue to firmware upgrade
    # enter FTP server IP and update image filename
    
    # Put swtich into admin mode so that we can update it.
    logger.log("Setting switch to admin mode")
    telnet.write('admin start\r')
    reList = []
    reList.append(re.compile('%s>' % userName))
    reList.append(re.compile('already'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == 0:
        logger.log("Switch in admin mode.")
    elif index == 1:
        telnet.write('admin cancel\ry\r')
        reList = []
        reList.append(re.compile('%s>' % userName))
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
        if index == 0:
            telnet.write('admin start\r')
        else:
            logger.log("Error! An unexpected timeout has occurred.")
            logger.log("Last message received:\n%s" % txt)
            telnet.close()
            lastError = ComDefs.GENERAL_UPDATE_ERROR
            return lastError
    elif index == -1:
        logger.log("Error! Unhandled timeout.")
        logger.log("Last message received:\n%s" % txt)
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
        
    # Fetch update image from external FTP server
    flashfilesplit = os.path.split(filename)
    flashfilename = flashfilesplit[1]
    logger.log("Retrieving firmware image file from FTP server.", ComDefs.TRUE)
    # Clean up any old firmware images so we ensure we have a good image.
    telnet.write('image cleanup\r')
    reList = []
    reList.append(re.compile('%s>' % userName))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message received:\n%s" % txt)
        telnet.write('admin stop\r')
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    
    # Grab the image
    logger.log('image fetch %s %s %s %s' % (params.ftpUserName, params.ftpIPAddress, filename, flashfilename))
    telnet.write('image fetch %s %s %s %s\r' % (params.ftpUserName, params.ftpIPAddress, filename, flashfilename))
    
    if params.ftpUserName != "anonymous":
        reList = []
        reList.append(re.compile('Password:'))
        reList.append(re.compile('ftp>'))
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)

        if index == 0:
            telnet.write("%s\r" % params.ftpPassword)
        elif index == -1:
            logger.log("Error! Telnet timeout.")
            logger.log("Last message received:\n%s" % txt)
            telnet.close()
            return ComDefs.GENERAL_UPDATE_ERROR

    reList = []
    reList.append(re.compile('Failed'))
    reList.append(re.compile('ftp>'))
    reList.append(re.compile('%s>' % userName))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT_LONGER)

    if index == 0:
        logger.log("Unable to retrieve firmware image file from FTP server.", ComDefs.TRUE)
        telnet.write('admin stop\r')
        lastError = ComDefs.IO_FIRMWARE_IMAGE_NOT_FOUND
        telnet.close()
        return lastError
    elif index == 1:
        telnet.write('quit\r')
        reList = []
        reList.append(re.compile('%s>' % userName))
        (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)

        if index == 0:
            telnet.write('admin stop\r')
            reList = []
            reList.append(re.compile('%s>' % userName))
            (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)

            if index == 0:
                telnet.write('exit\r')
            elif index == -1:
                logger.log("Error! Telnet timeout.")
                logger.log("Last message received:\n%s" % txt)
                telnet.close()
                return ComDefs.GENERAL_UPDATE_ERROR

        elif index == -1:
            logger.log("Error! Telnet timeout.")
            logger.log("Last message received:\n%s" % txt)
            telnet.close()
            return ComDefs.GENERAL_UPDATE_ERROR

        telnet.close()        
        lastError = ComDefs.FTP_INVALID_LOGIN
        return lastError
    elif index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message received:\n%s" % txt)
        telnet.write('admin stop\r')
        telnet.close()
        return ComDefs.GENERAL_UPDATE_ERROR
    
    
    # Update switch firmware
    logger.log("Updating Fibre Channel Switch Module firmware.", ComDefs.TRUE)
    telnet.write('image unpack %s\r' % flashfilename)
    reList = []
    reList.append(re.compile('switch\ reset'))
    reList.append(re.compile('file\ does\ not\ exist'))
    reList.append(re.compile('Error\ executing\ command'))
    reList.append(re.compile('BadSize'))
    reList.append(re.compile('CorruptImage'))
    reList.append(re.compile('rong\ QLogic\ image'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT_LONGER)

    if index == 1:
        logger.log("Error! Switch was unable to locate firmware flash file.")
        telnet.write('admin stop\r')
        lastError = ComDefs.IO_FIRMWARE_IMAGE_NOT_FOUND
        telnet.close()
        return lastError
    elif index == 2:
        logger.log("Error! An error occurred while executing the flash command.")
        telnet.write('admin stop\r')
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        telnet.close()
        return lastError
    elif index == 3:
        logger.log("Error!  The firmware image has an invalid size.")
        telnet.write('admin stop\r')
        lastError = ComDefs.IO_INVALID_IMAGE_FILE
        telnet.close()
        return lastError
    elif index == 4:
        logger.log("Error!  The firmware image is corrupted.")
        telnet.write('admin stop\r')
        lastError = ComDefs.IO_INVALID_IMAGE_FILE
        return lastError
    elif index == 5:
        logger.log("Error!  Wrong QLogic image for this switch.")
        telnet.write('admin stop\r')
        lastError = ComDefs.IO_INVALID_IMAGE_FILE
        return lastError
    elif index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message received:\n%s" % txt)
        telnet.write('admin stop\r')
        telnet.close()
        return ComDefs.GENERAL_UPDATE_ERROR

    telnet.write('\r')
    reList = []
    reList.append(re.compile('%s>' % userName))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    
    
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message received:\n%s" % txt)
        telnet.write('admin stop\r')
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError

    logger.log("Resetting the switch.")
    telnet.write('reset\r')
    reList = []
    reList.append(re.compile('[y/n]'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)

    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message received:\n%s" % txt)
        telnet.write('admin stop\r')
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
    
    telnet.write('y\r')
    
    reList = []
    reList.append(re.compile('Reset switch in progress'))
    (index,mo,txt) = telnet.expect(reList, ComDefs.TN_TIMEOUT)
    
    if index == -1:
        logger.log("Error! Telnet timeout.")
        logger.log("Last message received:\n%s" % txt)
        telnet.write('admin stop\r')
        telnet.close()
        lastError = ComDefs.GENERAL_UPDATE_ERROR
        return lastError
     
    logger.log("Fibre Channel Switch Module has been reset.", ComDefs.TRUE)
    if standalone:
        logger.log("Success! Fibre Channel Switch Module update complete.", ComDefs.TRUE)
    
    lastError = ComDefs.SUCCESS
    return lastError
    
def main():
    """ Main program entry point when running standalone """
    global params, logger, lastError, ipAddr
    params = ComDefs.parameters()
    standalone = ComDefs.TRUE
    
    # Parse command line options with getopt()
    try:
        opts, args = getopt.getopt(sys.argv[1:], "?hvVi:u:p:s:m:w:n:",
             ["?", "help", "verbose", "version", 
              "ipaddress=", "username=", "password=",
              "ftpserver=", "ftpusername=", "ftppassword=", "imagename="])

    except getopt.GetoptError:
        # Print usage information and exit
        usage()
        sys.exit(0)
    
    lastError = parse_args(opts)
    if lastError != ComDefs.SUCCESS:
        return lastError

    lastError = verify_args()
    if lastError != ComDefs.SUCCESS:
        return lastError
        
    logger = Logger(params)
    
    rc = flash_switch(ipAddr)
    if rc != ComDefs.SUCCESS:
        logger.log(ComDefs.GetLastError(rc), ComDefs.TRUE)
    
    
# This get's invoked if they run this script directly
if __name__ == '__main__':
    main()
